from . import TestBoolean
from . import TestChamfer
from . import TestDatum
from . import TestDraft
from . import TestFillet
from . import TestHelix
from . import TestHole
from . import TestInvoluteGear
from . import TestLinearPattern
from . import TestLoft
from . import TestMirrored
from . import TestMultiTransform
from . import TestPad
from . import TestPipe
from . import TestPocket
from . import TestPolarPattern
from . import TestPrimitive
from . import TestRevolve
from . import TestShapeBinder
from . import TestThickness
from . import TestTopologicalNamingProblem
