def create_nodes(femmesh):
    # nodes
    femmesh.addNode(0.0, 2000.0, 0.0, 1)
    femmesh.addNode(0.0, 0.0, 0.0, 2)
    femmesh.addNode(2000.0, 2000.0, 0.0, 3)
    femmesh.addNode(2000.0, 0.0, 0.0, 4)
    femmesh.addNode(4000.0, 2000.0, 0.0, 5)
    femmesh.addNode(4000.0, 0.0, 0.0, 6)
    femmesh.addNode(6000.0, 2000.0, 0.0, 7)
    femmesh.addNode(6000.0, 0.0, 0.0, 8)
    femmesh.addNode(8000.0, 2000.0, 0.0, 9)
    femmesh.addNode(8000.0, 0.0, 0.0, 10)
    femmesh.addNode(10000.0, 2000.0, 0.0, 11)
    femmesh.addNode(10000.0, 0.0, 0.0, 12)
    femmesh.addNode(12000.0, 2000.0, 0.0, 13)
    femmesh.addNode(12000.0, 0.0, 0.0, 14)
    femmesh.addNode(1000.0, 1000.0, 2000.0, 15)
    femmesh.addNode(3000.0, 1000.0, 2000.0, 16)
    femmesh.addNode(5000.0, 1000.0, 2000.0, 17)
    femmesh.addNode(7000.0, 1000.0, 2000.0, 18)
    femmesh.addNode(9000.0, 1000.0, 2000.0, 19)
    femmesh.addNode(11000.0, 1000.0, 2000.0, 20)
    return True


def create_elements(femmesh):
    # elements
    femmesh.addEdge([1, 2], 1)
    femmesh.addEdge([3, 4], 2)
    femmesh.addEdge([5, 6], 3)
    femmesh.addEdge([7, 8], 4)
    femmesh.addEdge([9, 10], 5)
    femmesh.addEdge([11, 12], 6)
    femmesh.addEdge([13, 14], 7)
    femmesh.addEdge([1, 3], 8)
    femmesh.addEdge([2, 4], 9)
    femmesh.addEdge([3, 5], 10)
    femmesh.addEdge([4, 6], 11)
    femmesh.addEdge([5, 7], 12)
    femmesh.addEdge([6, 8], 13)
    femmesh.addEdge([7, 9], 14)
    femmesh.addEdge([8, 10], 15)
    femmesh.addEdge([9, 11], 16)
    femmesh.addEdge([10, 12], 17)
    femmesh.addEdge([11, 13], 18)
    femmesh.addEdge([12, 14], 19)
    femmesh.addEdge([1, 15], 20)
    femmesh.addEdge([3, 16], 21)
    femmesh.addEdge([5, 17], 22)
    femmesh.addEdge([7, 18], 23)
    femmesh.addEdge([9, 19], 24)
    femmesh.addEdge([11, 20], 25)
    femmesh.addEdge([2, 15], 26)
    femmesh.addEdge([4, 16], 27)
    femmesh.addEdge([6, 17], 28)
    femmesh.addEdge([8, 18], 29)
    femmesh.addEdge([10, 19], 30)
    femmesh.addEdge([12, 20], 31)
    femmesh.addEdge([15, 3], 32)
    femmesh.addEdge([16, 5], 33)
    femmesh.addEdge([17, 7], 34)
    femmesh.addEdge([18, 9], 35)
    femmesh.addEdge([19, 11], 36)
    femmesh.addEdge([20, 13], 37)
    femmesh.addEdge([15, 4], 38)
    femmesh.addEdge([16, 6], 39)
    femmesh.addEdge([17, 8], 40)
    femmesh.addEdge([18, 10], 41)
    femmesh.addEdge([19, 12], 42)
    femmesh.addEdge([20, 14], 43)
    femmesh.addEdge([15, 16], 44)
    femmesh.addEdge([16, 17], 45)
    femmesh.addEdge([17, 18], 46)
    femmesh.addEdge([18, 19], 47)
    femmesh.addEdge([19, 20], 48)
    femmesh.addEdge([2, 3], 49)
    femmesh.addEdge([4, 5], 50)
    femmesh.addEdge([6, 7], 51)
    femmesh.addEdge([8, 9], 52)
    femmesh.addEdge([10, 11], 53)
    femmesh.addEdge([12, 13], 54)
    return True
